unit ICFileMapWhat;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This module implements the FileMap What, which is used to drive the entire
	File Mappings dialog.  The dialog is kinda complicated, so I guess the module
	is kinda complicated too.
	
	The bulk of the code falls into two chunks. One chunk (which comes second
	in the file) operates the File Mapping window itself, with the usual array of
	standard What functions and commands that implement each of the buttons.
	The other chunk (which comes first in the file) is responsible for driving
	the Add/Edit Map Entry modal dialog.
	
	The key data structures are a ICMapEntry list stored in the whatData field,
	and a ListManager ListHandle in the whatSpareData.  The ICMapEntry list
	is a complicated data structure, with many variable length fields.
	Fortunately, the IC API provides a bunch of useful functions for dealing
	with this data structure. The ListManger list is only used for display
	and editing purposes, and no data is actually stored in the list.
	
	The module implements its own 'LDEF' to draw the list. This is done
	using a neat trick, namely the SimpleLDEF. This LDEF is a stub that
	assumes that the list's refCon field points to a procedure that
	implements the real LDEF, and simply routes all LDEF calls through
	to the refCon.
	
	One standard gotcha throughout this module is that List Manager
	lists are zero based, and the ICMapEntry list is one based.  Don't you just
	hate it when two worlds collide?
	
	Another weird thing is that we pass an "entry" parameter around to
	lots of routines when we don't strictly need to. For example,
	DoAddChangeButtons. This is because we're trying to avoid
	allocating two ICMapEntries because they're *huge*.  So the
	original calling routine allocates the entry and passes it around
	to the various routines that it calls, even though it doesn't
	need to pass any information in or out.  Routines that do this
	are marked with the special phrase
	
		"See comment about entry parameter at top of this file."
		
	Hack warning: This What needs to implement a series of Defaults
	buttons to set the File Mapping to the default value.  This used
	to be easy because the default mapping was in the applications
	resource fork.  Now that the component sets up the default value,
	the application needs to get it from somewhere.  After much
	thought about adding an API routine for this, I decided to get it
	from the extension in the Extensions folder.  See the routine
	GetDefaultMapEntries for more details.  As with all places
	where the application has intimate knowledge of the IC
	component, I hope to get rid of this eventually.  -- Quinn, IC 1.4
*)

interface

	uses
		ICWindowGlobals;

	(* ***** Standard What Interface ***** *)
	
	(* See the comments in ICWindowGlobals. *)

	function WhatInitFileMap(dummy : integer) : OSStatus;
	function WhatOpenFileMap (wt: WindowType; item: integer): OSStatus;
	function WhatFlushFileMap (wt: WindowType; item: integer): OSStatus;
	function WhatCloseFileMap (wt: WindowType; item: integer): OSStatus;
	function WhatKeyFileMap (wt: WindowType; item: integer; event: EventRecord): OSStatus;
	function WhatClickFileMap (wt: WindowType; item: integer; event: EventRecord): OSStatus;
	function WhatActivateFileMap (wt: WindowType; item: integer; activate: Boolean): OSStatus;
	function WhatCursorFileMap (wt: WindowType; item: integer; pt: Point; cursorID: integer): OSStatus;

	(* ***** Layer Breaking Interface ***** *)

	procedure MappingSetSelection(key:Str255);
		(* Called by the FindWindowAndSelectKey routine in ICEditPrefAppleEvents.
			This routine is called to select the entry in the File Mapping list that
			is best described by the key. In this case the key is the file mapping's
			extension field.
			
			For example, if an Edit Preference AppleEvent is sent with
			the parameter "Mapping.txt", the FindWindowAndSelectKey matches
			and strips off the "Mapping", and passes ".txt" in as the key parameter
			to this routine.
		*) 
	
implementation

	uses
		Icons,
		Dialogs,
		Lists,
		Fonts,
		ToolUtils,
		Resources,
		
		InternetConfig, 
		
		ICResources,
		ICStringLists,
		ICDialogs,
		ICMiscSubs,
		ICCommonSubs,
		ICWhatSubs,
		ICGlobals,
		ICIconSuites,
		ICStandardFile,
		ICMovableModal,
		ICWindowGlobals,
		ICInstall;

	var
		gListUserItemUPP : UserItemUPP;
		gFileMapLDEFUPP : ListDefUPP;
	
	const
		kMappingCellHeight = kListCellHeight;
		(* kMappingCellHeight is the height of the cells in the File Map list. *)

	(* ***** Utilities ***** *)

	function GetDefaultMapEntries(var defaultEntries:Handle):OSStatus;
		(* This routine gets the default mapping resource out of
			the Internet Config Extension's resource fork. It's
			used to support the various "Default" buttons. Note
			that defaultEntries is returned as a memory handle.
		*)
		var
			err : OSStatus;
			extensionFileSpec : FSSpec;
			extensionResourceRefNum : integer;
	begin
		(* First check whether the extension is installed.  IsExtensionInstalled
			returns true if it is, and also sets extensionFileSpec to its location.
			If we can't find the extension, chances are we're running without
			component manager, in which case we return unimpErr so that
			the error handler tells the user they system software is
			insufficiently cool.
		*)
		defaultEntries := nil;
		err := noErr;
		if not IsExtensionInstalled(extensionFileSpec) then begin
			err := fnfErr;
		end; (* if *)
		
		(* Now open up the extension's resourc fork. *)
		if err = noErr then begin
			extensionResourceRefNum := FSpOpenResFile(extensionFileSpec, fsRdPerm);
			err := ResError;
			if err = noErr then begin
			
				(* Get the default mapping database out of the extension
					and detach it, so that it's not cleaned up when we close
					the extension's resource fork.
				*)
				defaultEntries := Get1NamedResource('Pref', kICMapping);
				err := ResError;
				if (defaultEntries = nil) & (err = noErr) then begin
					err := resNotFound;
				end; (* if *)
				if err = noErr then begin
					DetachResource(defaultEntries);
				end; (* if *)
				
				(* Clean up. *)
				CloseResFile(extensionResourceRefNum);
			end; (* if *)

		end; (* if *)
		GetDefaultMapEntries := err;
	end; (* GetDefaultMapEntries *)

	(* ***** Edit Map Entry Modal Dialog ***** *)

	procedure SetFileFormatRadioButtons (modal: DialogPtr; itemHit: integer);
		(* This routine turns on the radio button associated with itemHit
			and turns off all the others, thereby maintaining the great radio
			button invariant. It's here, rather than nested inside DoAddChangeButtons,
			because it's called by AddChangeFilter.
		*)
	begin
		SetDControlBoolean(modal, ditEditMapAsciiRadio, itemHit = ditEditMapAsciiRadio);
		SetDControlBoolean(modal, ditEditMapBinaryRadio, itemHit = ditEditMapBinaryRadio);
		SetDControlBoolean(modal, ditEditMapMacintoshRadio, itemHit = ditEditMapMacintoshRadio);
	end; (* SetFileFormatRadioButtons *)

	function AddChangeFilter (dlg: DialogPtr; var event: EventRecord; var item: integer): Boolean;
		(* This routine is the modal dialog filter function for the Add/Edit Map
			Entry modal dialog. It calls through to the standard OKCancelModalFilter
			routine (which handles return, enter, command-., and so on)
			and also handles command-1, 2 and 3 for setting the file format
			radio buttons. Note that this routine is not nested inside
			DoAddChangeModalDialog because filter functions have to be
			global.
		*)
		var
			result: Boolean;
			charTyped: char;
	begin
		result := OKCancelModalFilter(dlg, event, item);
		if not result then begin
			if (event.what = keyDown) and (band(event.modifiers, cmdKey) <> 0) then begin
				charTyped := char(band(event.message, charCodeMask));
				if charTyped in ['1'..'3'] then begin
					SetFileFormatRadioButtons(dlg, ord(charTyped) - ord('1') + ditEditMapAsciiRadio);
				end; (* if *)
			end; (* if *)
		end; (* if *)
		AddChangeFilter := result;
	end; (* AddChangeFilter *)

	function DoAddChangeModalDialog (var entryToEdit: ICMapEntry): OSStatus;
		(* This routine runs the Add/Edit Map Entry modal dialog. It has many
			subroutines that are nested inside to indicate that they apply
			only to the Add/Edit Map Entry, and not to the File Mapping What
			as a whole. Also the entryToEdit parameter and the addChangeModalDialog
			are considered global variables from the perspective of the routines nested
			inside.
		*)

		(* This is a pseudo global variable that's a pointer to the Add/Edit
			Map Entry modal dialog.
		*)
		var
			addChangeModalDialog: DialogPtr;

		(* ***** Utilities for the Add/Edit Map Entry Modal Dialog ***** *)
		
		function MapOSTypeToString (anOSType: OSType): Str15;
			(* This routine returns anOSType as a string. It's a simple
				wrapper around OSTypeToString, but note the special
				handling of the null OSType. This is because IC
				uses the null OSType to signify that an entry hasn't
				ever been set. We don't want such an entry to
				come up as four null characters, because that will
				confuse the hell out of our users. So instead we
				need special processing.
			*)
		begin
			if anOSType = OSType(0) then begin
				MapOSTypeToString := '';
			end else begin
				MapOSTypeToString := OSTypeToString(anOSType);
			end; (* if *)
		end; (* MapOSTypeToString *)

		function MapStringToOSType (str: Str255): OSType;
			(* This routine is a simple wrapper around StringToOSType
				except that we append four spaces to the string
				before converting it. StringToOSType expects the
				source string to be four characters or more.  Appending
				four spaces ensures strings shorter than four characters
				will be converted consistently, and in a more user
				friendly fashion.
			*)
		begin
			str := TPCopy(concat(str, '    '), 1, 4);
			MapStringToOSType := StringToOSType(str);
		end; (* MapStringToOSType *)

		procedure UnpackPostProcessingInfo;
			(* This routine sets the post processing fields of the dialog
				up to match the fields in the entryToEdit record. This is really
				part of the SetAllTheStuff routine, except it's been abstracted
				out because it's used elsewhere.
			*)
		begin
			if entryToEdit.post_creator <> OSType(0) then begin
				SetDControlTitle(addChangeModalDialog, ditEditMapPostButton, concat(GetIndStr(rMiscStrings, strLeftSmartQuote), entryToEdit.post_app_name, GetIndStr(rMiscStrings, strRightSmartQuote)));
				SetDItemText(addChangeModalDialog, ditEditMapPostCreator, MapOSTypeToString(entryToEdit.post_creator));
			end else begin
				SetDControlTitle(addChangeModalDialog, ditEditMapPostButton, GetIndStr(rMiscStrings, strNone));
				SetDItemText(addChangeModalDialog, ditEditMapPostCreator, '');
			end; (* if *)
			SetDControlBoolean(addChangeModalDialog, ditEditMapPostCheckbox, btst(entryToEdit.flags, ICmap_post_bit));
		end; (* UnpackPostProcessingInfo *)

		procedure UnpackEntryIntoDialog;
			(* This routine sets all of the fields of the dialog up to match
				the entryToEdit record. It's called when the dialog is opened
				in order to set up the fields initially, and also when the
				user clicks the Defaults button, to sync up the fields with
				the new entry.
			*)
		begin
			(* First the boring text fields. *)
			SetDItemText(addChangeModalDialog, ditEditMapEntryName, entryToEdit.entry_name);
			SetDItemText(addChangeModalDialog, ditEditMapExtension, entryToEdit.extension);
			SetDItemText(addChangeModalDialog, ditEditMapMIMEType, entryToEdit.MIME_type);
			
			(* Now the file format radio buttons. *)
			SetDControlBoolean(addChangeModalDialog, ditEditMapAsciiRadio,
													not btst(entryToEdit.flags, ICmap_binary_bit));
			SetDControlBoolean(addChangeModalDialog, ditEditMapBinaryRadio,
													btst(entryToEdit.flags, ICmap_binary_bit) 
													and not (btst(entryToEdit.flags, ICmap_resource_fork_bit)));
			SetDControlBoolean(addChangeModalDialog, ditEditMapMacintoshRadio, 
													btst(entryToEdit.flags, ICmap_binary_bit) 
													and (btst(entryToEdit.flags, ICmap_resource_fork_bit)));
			
			(* Now the incoming/outgoing checkboxen. *)
			SetDControlBoolean(addChangeModalDialog, ditEditMapNotForIncoming, btst(entryToEdit.flags, ICmap_not_incoming_bit));
			SetDControlBoolean(addChangeModalDialog, ditEditMapNotForOutgoing, btst(entryToEdit.flags, ICmap_not_outgoing_bit));
			
			(* Now the file type and creator info. *)
			SetDItemText(addChangeModalDialog, ditEditMapType, MapOSTypeToString(entryToEdit.file_type));
			SetDItemText(addChangeModalDialog, ditEditMapCreator, MapOSTypeToString(entryToEdit.file_creator));
			SetDItemText(addChangeModalDialog, ditEditMapApplicationName, entryToEdit.creator_app_name);

			(* Now the post-processing info. *)
			UnpackPostProcessingInfo;
			
			(* Now select the first field. *)
			SelectDialogItemText(addChangeModalDialog, ditEditMapEntryName, 0, 32767);
		end; (* UnpackEntryIntoDialog *)

		procedure PackDialogIntoEntry;
			(* This routine is the logical opposite of UnpackEntryIntoDialog, in
				that it extracts the information out of the various fields in the
				dialog and puts it into the entryToEdit.
			*)
			const
				kKnownConstantsMask = ICmap_binary_mask +
														ICmap_data_fork_mask + 
														ICmap_resource_fork_mask + 
														ICmap_not_incoming_mask + 
														ICmap_not_outgoing_mask;
			var
				flags : longint;
				tempStr : Str255;
		begin
			(* First the boring text fields. *)
			GetDItemText(addChangeModalDialog, ditEditMapEntryName, entryToEdit.entry_name);
			GetDItemText(addChangeModalDialog, ditEditMapExtension, entryToEdit.extension);
			GetDItemText(addChangeModalDialog, ditEditMapMIMEType, entryToEdit.MIME_type);

			(* Now the file format radio buttons. *)
			flags := band(entryToEdit.flags, bnot(kKnownConstantsMask));
			flags := flags + ICmap_data_fork_mask;
			if not GetDControlBoolean(addChangeModalDialog, ditEditMapAsciiRadio) then begin
				flags := flags + ICmap_binary_mask;
			end; (* if *)
			if GetDControlBoolean(addChangeModalDialog, ditEditMapMacintoshRadio) then begin
				flags := flags + ICmap_resource_fork_mask;
			end; (* if *)

			(* Now the incoming/outgoing checkboxen. *)
			if GetDControlBoolean(addChangeModalDialog, ditEditMapNotForIncoming) then begin
				flags := flags + ICmap_not_incoming_mask;
			end; (* if *)
			if GetDControlBoolean(addChangeModalDialog, ditEditMapNotForOutgoing) then begin
				flags := flags + ICmap_not_outgoing_mask;
			end; (* if *)
			entryToEdit.flags := flags;

			(* Now the file type and creator info. *)
			GetDItemText(addChangeModalDialog, ditEditMapType, tempStr);
			entryToEdit.file_type := MapStringToOSType(tempStr);
			GetDItemText(addChangeModalDialog, ditEditMapCreator, tempStr);
			entryToEdit.file_creator := MapStringToOSType(tempStr);
			GetDItemText(addChangeModalDialog, ditEditMapApplicationName, entryToEdit.creator_app_name);

			(* At this point you'd expect us to extract the post-processing info.
				Oh well.  The post-processing info is kinda strange. Because
				we have some invisible state (namely that the post-processing
				application has never been set), we don't extract the post-processing
				info from the dialog fields. Instead the routines that operate
				on the post-processing info keep the entryToEdit fields
				and the dialog fields in sync.
			*)
		end; (* PackDialogIntoEntry *)
		
		(* ***** Command Buttons ***** *)

		procedure DoChooseExampleButton;
			(* This routine responds to clicks in the Choose Example button.
				The response is to put up a Standard File dialog asking for
				the example file, and place the resulting file into the fields
				of the Add/Edit Map Entry dialog.
			*)
			var
				err: OSStatus;
				infoForChosenFile: FInfo;
				junkFileSpec : FSSpec;
				chosenApplicationFileSpec: FSSpec;
		begin
			(* Ask the user for the example file. *)
			err := ICStandardGetFile(OSType(0), junkFileSpec, infoForChosenFile);
			DisplayError(acGetExample, err);
			if err = noErr then begin
				(* Fill in the File Type and Creator fields. *)
				SetDItemText(addChangeModalDialog, ditEditMapType,
															MapOSTypeToString(infoForChosenFile.fdType));
				SetDItemText(addChangeModalDialog, ditEditMapCreator,
															MapOSTypeToString(infoForChosenFile.fdCreator));
				
				(* Try to find the application that corresponds to the creator and,
					if you find it, put it into the application name field.
				*)
				err := FindApplicationInDTDB(infoForChosenFile.fdCreator, chosenApplicationFileSpec);
				if err = noErr then begin
					SetDItemText(addChangeModalDialog, ditEditMapApplicationName, chosenApplicationFileSpec.name);
				end else begin
					SetDItemText(addChangeModalDialog, ditEditMapApplicationName, '');
				end; (* if *)
			end; (* if *)
		end; (* DoChooseExampleButton *)

		function DoChoosePostProcessingApplication: Boolean;
			(* This routine prompts the user to select a post-processing
				application, and fills out the dialog fields to suit. It returns
				true if an application was chosen.
			*)
			var
				err: OSStatus;
				infoForChosenApplication: FInfo;
				chosenApplicationFileSpec: FSSpec;
		begin
			err := ICStandardGetFile('APPL', chosenApplicationFileSpec, infoForChosenApplication);
			DisplayError(acGetExample, err);
			if err = noErr then begin
				entryToEdit.post_creator := infoForChosenApplication.fdCreator;
				entryToEdit.post_app_name := chosenApplicationFileSpec.name;
			end; (* if *)
			DoChoosePostProcessingApplication := (err = noErr);
		end; (* DoChoosePostProcessingApplication *)

		procedure DoAddEditDefaultsButton;
			(* This routine handles the Defaults button in the Add/Edit Mapping Entry
				modal dialog (which is distinct from the Defaults button in the main File
				Mapping dialog). The basic idea is to look up the extension in the default
				mapping database and overwrite the existing entry based on the new
				entry.  Note that we don't dispose of the defaultEntries handle because
				it's just a resource and we might as well leave it lying around.
			*)
			var
				err:OSStatus;
				extension:Str255;
				defaultEntries:Handle;
				newEntry: ICMapEntry;
				junk: integer;
		begin
			(* Find the current extension in the default file mapping database. *)
			GetDItemText(addChangeModalDialog, ditEditMapExtension, extension);
			err := GetDefaultMapEntries(defaultEntries);
			if err = noErr then begin
				err := ICMapEntriesFilename(WhatGetDocumentInstance, defaultEntries, extension, newEntry);
				DisposeHandle(defaultEntries);
			end; (* if *)
			
			(* If all is OK, use the new entry. *)
			if err = noErr then begin
				entryToEdit := newEntry;
				UnpackEntryIntoDialog;
			end; (* if *)
			
			if err = icPrefNotFoundErr then begin
				(* Special case this error, putting up a dialog that says that
					we don't know what to do for this extension.
				*)
				ParamText('', extension, '', '');
				junk := StopAlert(rPrefNotFoundAlert, gOKModalFilterUPP);
			end else begin
				DisplayError(acSetDefaults, err);
			end; (* if *)
		end; (* DoAddEditDefaultsButton *)
		
		var
			err: OSStatus;
			hitItem: integer;
			junkBool: Boolean;
			movableModalSavedMenuBarState: Ptr;
	begin
		(* Create the modal dialog. *)
		err := noErr;
		addChangeModalDialog := GetNewDialog(rEditMapEntryDialog, nil, WindowPtr(-1));
		if addChangeModalDialog = nil then begin
			err := memFullErr;
		end else begin
			SetupDefaultButtonUserItem(addChangeModalDialog, ditOK, ditEditMapDefaultUserItem);
			UnpackEntryIntoDialog;
			ShowWindow(addChangeModalDialog);
			DisableMenuBar(movableModalSavedMenuBarState);
			InitCursor;
			
			(* Run the modal dialog, handling each item hit. *)
			repeat
				MovableModalDialog(@AddChangeFilter, hitItem);
				case hitItem of
					ditEditMapAsciiRadio..ditEditMapMacintoshRadio:
						SetFileFormatRadioButtons(addChangeModalDialog, hitItem);
					ditEditMapNotForIncoming..ditEditMapNotForOutgoing:
						ToggleDControlBoolean(addChangeModalDialog, hitItem);
					ditEditMapChooseExample:
						DoChooseExampleButton;
					ditEditMapDefaults:
						DoAddEditDefaultsButton;

					ditEditMapPostCheckbox:  begin
						(* Special logic for the Post Processing checkbox. If it's currently
							set then just turn it off. It it's currently clear and there's never
							been a post-processing application set, then clicking the checkbox
							asks the user to specify one using DoChoosePostProcessingApplication.
						*)
						if not btst(entryToEdit.flags, ICmap_post_bit) then begin
							if (entryToEdit.post_creator = OSType(0)) then begin
								if entryToEdit.file_creator <> OSType(0) then begin
									entryToEdit.post_creator := entryToEdit.file_creator;
									entryToEdit.post_app_name := entryToEdit.creator_app_name;
								end else begin
									junkBool := DoChoosePostProcessingApplication;
								end; (* if *)
							end; (* if *)
							if entryToEdit.post_creator <> OSType(0) then begin
								bset(entryToEdit.flags, ICmap_post_bit);
							end; (* if *)
						end else begin
							bclr(entryToEdit.flags, ICmap_post_bit);
						end; (* if *)
						
						(* Finally, call UnpackPostProcessingInfo to sync the
							dialog up with our changes to entryToEdit.
						*)
						UnpackPostProcessingInfo;
					end;

					ditEditMapPostButton:  begin
						if DoChoosePostProcessingApplication then begin
							bset(entryToEdit.flags, ICmap_post_bit);
						end; (* if *)
						UnpackPostProcessingInfo;
					end;

					otherwise
						(* do nothing *) ;
				end; (* case *)
			until hitItem in [ditOK, ditCancel];
			
			(* Clean up. *)
			if hitItem = ditOK then begin
				PackDialogIntoEntry;
			end else begin
				err := userCanceledErr;
			end; (* if *)
			ReEnableMenuBar(movableModalSavedMenuBarState);
			DisposeDialog(addChangeModalDialog);
		end; (* if *)
		DoAddChangeModalDialog := err;
	end; (* DoAddChangeModalDialog *)

	(* ***** Utilities for the File Mappings Window ***** *)

	procedure GetEntryName (listH: ListHandle; listCell: Cell; var result: Str255);
		(* Returns the name associated with the cell in the ListHandle.
			This is used by the LDoKey routine to support selection by
			typing into the list.
		*)
		var
			entries: Handle;
			entry: ICMapEntry;
			junkPos: longint;
	begin
		result := '';
		entries := Handle(gWindows[MapWindowPtrToWindowType(listH^^.port)].items[ditFileMapList]^.whatData);
		if ICGetIndMapEntry(WhatGetDocumentInstance, entries, listCell.v + 1, junkPos, entry) = noErr then begin
			result := entry.extension;
		end; (* if *)
	end; (* GetEntryName *)

	function SetEntriesHandle(wt:WindowType; newEntries:Handle):OSStatus;
		(* Sets the global entries list to newEntries, and adjusts the user
			interface to suit.  The basic approach is to count the number of
			existing entries, delete all the rows in the list, and then
			add the count of the number of entries back into the list.
			Regardless of whether it returns an error, this routine 
			disposes of the newEntries handle.
		*)
		var
			err:OSStatus;
			listH:ListHandle;
			tmpH:Handle;
			countOfNewEntries:longint;
			junkRow:integer;
	begin
		listH := ListHandle(gWindows[wt].items[ditFileMapList]^.whatSpareData);
		err := ICCountMapEntries(WhatGetDocumentInstance, newEntries, countOfNewEntries);
		if err = noErr then begin
		
			(* Swap newEntries and the global entries list (in whatData). *)
			tmpH := newEntries;
			newEntries := Handle(gWindows[wt].items[ditFileMapList]^.whatData);
			gWindows[wt].items[ditFileMapList]^.whatData := Ptr(tmpH);

			(* Delete the existing rows and add in the new rows. *)
			LDelRow(listH^^.dataBounds.bottom + 10, 0, listH);
			junkRow := LAddRow(countOfNewEntries , 0, listH);
		end; (* if *)
		DisposeHandle(newEntries);
		SetEntriesHandle := err;
	end; (* SetEntriesHandle *)

	procedure DimButtons (wt: WindowType);
		(* Dims/undims the buttons in the dialog depending on the selection. This
			routine is typically called after the selection has been changed in some
			way.
		*)
		var
			listH: ListHandle;
			window : WindowPtr;
			haveSelection: Boolean;
	begin
		listH := ListHandle(gWindows[wt].items[ditFileMapList]^.whatSpareData);
		window := gWindows[wt].window;
		haveSelection := (LSelectedLine(listH) <> -1);
		SetDControlEnable(window, ditFileMapAdd, not WhatIsLocked(wt, ditFileMapList));
		SetDControlEnable(window, ditFileMapDelete, haveSelection and not WhatIsLocked(wt, ditFileMapList));
		SetDControlEnable(window, ditFileMapChange, haveSelection and not WhatIsLocked(wt, ditFileMapList));
		SetDControlEnable(window, ditFileMapDefaults, not WhatIsLocked(wt, ditFileMapList));
		DefaultButtonUserItem(window, ditFileMapDefaultUserItem);
	end; (* DimButtons *)

	(* ***** Drawing ***** *)

	procedure FileMapLDEF (message: integer; drawSelected: Boolean;
											var theCellRect: Rect;
											theCell: Cell;
											dataOffset: integer; dataLength: integer; 
											listH: ListHandle);
		(* This LDEF is responsible for drawing the main list item in the File Mappings
			dialog. It basically is a simple dispatcher based on message type, where the
			only message types we're interested in are the draw and highlight message.
		*)
		
		procedure FileMapLDEFDraw(entries: Handle);
			(* This is the core code for the LDEF, which draws a list cell.
				It's stupid for me to describe what the drawing looks like here.
				Run the application and take a look.

				Oh yeah, the + 0 in some of these equations is deliberate. This
				used to be a fudge factor in the drawing, which I've since taken
				out. I've left it in the code because I may need to put the fudge back
				in some day.
			*)
			var
				err: OSStatus;
				entry: ICMapEntry;
				junkPos: longint;
				currentXOffset: integer;
				systemFontInfo : FontInfo;
				applicationFontInfo: FontInfo;
				offsetToTopOfFirstTextLine: integer;
				iconRegion: RgnHandle;
		begin
			if dataLength = 0 then begin

				(* Get the data associated with this entry. *)
				err := ICGetIndMapEntry(WhatGetDocumentInstance, entries, theCell.v + 1, junkPos, entry);
				if err = noErr then begin

					(* Clean the cell to start with. *)
					EraseRect(theCellRect);
					
					(* Plot the icon, if we have the system software to do so. Maintain
						iconRegion as the region that was plotted, so we can exclude it
						from the highlighting. Also maintain the current X offset, so
						we can draw the text flush with the left hand side if we didn't
						plot an icon.
					*)
					currentXOffset := 0;
					iconRegion := NewRgn;
					if iconRegion <> nil then begin
						DrawListIcon(entry.file_type, entry.file_creator,
												drawSelected, theCellRect,
												currentXOffset, iconRegion);
					end; (* if *)
					currentXOffset := currentXOffset + 2;

					(* Collect information about the fonts we're using. *)
					TextFont(systemFont);
					GetFontInfo(systemFontInfo);
					TextFont(applFont);
					GetFontInfo(applicationFontInfo);

					(* Draw the two lines of text. *)
					offsetToTopOfFirstTextLine := (kMappingCellHeight - 
																				(	systemFontInfo.ascent +
																					systemFontInfo.descent + 
																					0 + 
																					applicationFontInfo.ascent + 
																					applicationFontInfo.descent)
																				) div 2;
					TextFont(systemFont);
					MoveTo(theCellRect.left + currentXOffset, theCellRect.top + 
																						offsetToTopOfFirstTextLine + 
																						systemFontInfo.ascent);
					DrawString(concat(entry.creator_app_name, 
													GetIndStr(rMiscStrings, strLongDash), 
													entry.entry_name));
					TextFont(applFont);
					MoveTo(theCellRect.left + currentXOffset, theCellRect.top + 
																						offsetToTopOfFirstTextLine + 
																						systemFontInfo.ascent + 
																						systemFontInfo.descent + 
																						0 + 
																						applicationFontInfo.ascent);
					DrawString(entry.extension);

					(* If the cell is selected, draw the highlight region. *)
					if drawSelected then begin
						HighlightListCell(theCellRect, iconRegion);
					end; (* if *)
					if iconRegion <> nil then begin
						DisposeRgn(iconRegion);
					end; (* if *)
				end; (* if *)
			end; (* if *)
		end; (* FileMapLDEFDraw *)

		var
			entries : Handle;
	begin
		{$unused dataOffset}
		SetPort(listH^^.port);
		entries := Handle(gWindows[MapWindowPtrToWindowType(listH^^.port)].items[ditFileMapList]^.whatData);
		case message of
			lInitMsg: 
				(* do nothing *) ;
			lDrawMsg,
			lHiliteMsg: 
				FileMapLDEFDraw(entries);
			lCloseMsg: 
				(* do nothing *) ;
		otherwise
			(* do nothing *) ;
		end; (* case *)
	end; (* FileMapLDEF *)

	(* ***** Add / Change Buttons ***** *)

	function DoAddChangeButtons (itemToChange: integer; var entry: ICMapEntry;	
															entries: Handle; listH: ListHandle): OSStatus;
		(* This routine is called in response to a request to change a list element.
			Its basic pupose is to put up the Add/Edit Mapping modal dialog.
			itemToChange is the list item number (ie zero based) of the item
			to edit. If itemToChange is -1, we're adding an item.
			See comment about entry parameter at top of this file.
		*)
		var
			err: OSStatus;
			junkRow: integer;
			posOfItemToChange: longint;
	begin
		(* If we're adding a new item, create a dummy record, otherwise
			fetch the entry to change from the entries list.
		*)
		if itemToChange = -1 then begin
			entry.version := 0;
			entry.file_type := OSType(0);
			entry.file_creator := OSType(0);
			entry.post_creator := OSType(0);
			entry.flags := 0;
			entry.extension := '';
			entry.creator_app_name := '';
			entry.post_app_name := '';
			entry.MIME_type := '';
			entry.entry_name := '';
			err := noErr;
		end else begin
			err := ICGetIndMapEntry(WhatGetDocumentInstance, entries, itemToChange + 1, 
														posOfItemToChange, entry);
		end; (* if *)

		(* Pose the add/edit modal dialog. *)
		if err = noErr then begin
			err := DoAddChangeModalDialog(entry);
		end; (* if *)

		if err = noErr then begin

			(* Everything is cool, write the new entry back to the entries handle
				If we're creating a new entry, append it to the end of the list,
				othewise overwrite the current entry.
			*)
			if itemToChange = -1 then begin
				err := ICAddMapEntry(WhatGetDocumentInstance, entries, entry);
				if err = noErr then begin
					junkRow := LAddRow(1, 32767, listH);
				end; (* if *)
			end else begin
				err := ICSetMapEntry(WhatGetDocumentInstance, entries, posOfItemToChange, entry);
			end; (* if *)
		end; (* if *)

		DoAddChangeButtons := err;
	end; (* DoAddChangeButtons *)

	(* ***** Sort Buttons ***** *)
	
	(* This code implements the sort buttons in the File Mappings window.
		These sort buttons are unusual in that they only sort the list as
		it stands; IC doesn't maintain the chosen sort order. We did this
		because we're slack. I hope to change this in some future version,
		but for the moment it'll stand.
		
		The sorting approach is that we copy the relevant string (ie
		the sort key) into a temporary array of SmallRecords, then sort those
		SmallRecords using QuickSort, then use the sorted list to build a new
		sorted array of entries.  The use of SmallRecords is good because it reduces
		the temporary work space used, and the amount of copying we have to do
		when we swap records (which we do a lot).
		
		A word about stack space: QuickSort has a pathologically
		bad case when the data is already sorted. This results in it generating
		stack frames equal to the number of items to be sorted. This can very
		quickly exceed the Mac's stack. IC has a very large stack, but it's still
		a worry. I've put stack space check in the recursive routine, just to be sure.
		
		Some obvious improvements to this would be:
		
		1. Maintain the list sorted, and record what the current sort order
			is. This would be a much better human interface.

		2. Use Insertion Sort for the QuickSorts of small number of items,
			because generally it has lower overhead for small lists.

		3.	Figure out some way to improve the speed of string comparisons.
			At the moment, this sorting is really slow, and the primary
			reason for that is IUCompString. When I substituted RelString,
			it sped up by a factor of 15. RelString is also more appropriate
			to use for things like extensions and the MIME type. However
			for other things, it's not so cool.  Unfortunately the sort order
			needs to be tied in with the LDoKey routine, which uses
			IUCompString. When I revisit this code, I'll most probably
			switch to RelString for most fileds, and rework the LDoKey
			to accept a comparison function pointer
			
		4.	QuickSort is pathologically bad (ie O(n^2)) if the data is already
			sorted. It would be way smart to pre-parse the list to see if its
			close to sorted and, if so, use Insertion Sort instead.
		
		5. I should also investigate sorting algorithms to see if there's
			one that I can use that minimises the number of comparisons,
			even at the expense of swaps. Given the slow IUCompString,
			it would pay to use a bucket more swaps if I can get away with
			it.
		
		6. If I can't make the sorting significantly faster, I need to put cursor
			spinning code in here.
	*)
	
	type
		SmallRecord =
			record
				position:longint;
				key:Str31;
			end;
		SmallArray = array[1..10000] of SmallRecord;
		SmallArrayPtr = ^SmallArray;
		SmallArrayHandle = ^SmallArrayPtr;
	
	var
		SmallSort_v : SmallRecord;
		SmallSort_t : SmallRecord;
		SmallSort_small:SmallArrayPtr;
		(* I took these variables out of QuickSort to reduce the stack usage.  I also made
			 them globals to make them go faster.
		*)
		
	function SmallSort(smallArray : SmallArrayHandle; count : longint) : OSStatus;
		(* Sort the small array whose size is count. If the routine returns
			an error, smallArray is most probably in a bad way, ie not sorted
			and not in its original state.
		*)
	
		label
			666;
	
		procedure QuickSort(l, r : longint);
			(* This code was typed in verbatim from Sedgewick's "Algorithms" book,
				except the modification as noted below.
			*)
			var
				i, j:longint;
		begin
			if r > l then begin
				(* Sedgewick always uses the r element for SmallSort_v,
					which is really bad if the list is sorted.  Instead we choose the
					element midway between l and r, which gives good behaviour
					if the list is sorted.  Note that we have to swap this element
					with the r'th element so that the right of the algorithm works.
				*)
				i := (l + r) div 2;
				SmallSort_v := SmallSort_small^[ i ];
				SmallSort_small^[ i ] := SmallSort_small^[ r ];
				SmallSort_small^[ r ] := SmallSort_v;
				(* SmallSort_v := SmallSort_small^[ r ]; *)

				(* Now continue on with the Sedgewick algorithm verbatim. *)
				i := l - 1;
				j := r;
				repeat
					repeat
						i := i + 1;
					until IUCompString(SmallSort_small^[i].key, SmallSort_v.key) >= 0;
					repeat
						j := j - 1;
					until  IUCompString(SmallSort_small^[j].key, SmallSort_v.key) <= 0;
					SmallSort_t := SmallSort_small^[i];
					SmallSort_small^[i] := SmallSort_small^[j];
					SmallSort_small^[j] := SmallSort_t;
				until j <= i;
				SmallSort_small^[j] := SmallSort_small^[i];
				SmallSort_small^[i] := SmallSort_small^[r];
				SmallSort_small^[r] := SmallSort_t;
				if StackSpace < 4096 then begin			(* See comment "A word about stack space" above. *)
					goto 666;
				end; (* if *)
				QuickSort(l, i-1);
				QuickSort(i+1, r);
			end; (* if *)
		end; (* QuickSort *)
		
		var
			s : SInt8;
	begin
		s := HGetState(Handle(smallArray));
		HLock(Handle(smallArray));
		SmallSort_small := smallArray^;
		
		(* This odd logic is because of a bug in the CodeWarrior Pascal compiler,
			that was causing my original more sensible logic to crash horribly.
			As the resulting workaround isn't particularly ugly, I haven't bothered
			to investigate whether the original bug has been fixed.
		*)
		SmallSort := memFullErr;
		QuickSort(1, count);
		SmallSort := noErr;
	666:
		HSetState(Handle(smallArray), s);
	end; (* SmallSort *)
	
	procedure DoSortButtons(wt: WindowType; item:integer; var entry : ICMapEntry);
		(* Implements the sort buttons in the File Mapping window. It sorts
			the file mapping list based on the key specified by the smallArray,
			ie if the item is ditFileMapSortExtension it sorts by extension, and so on.
			See comment about entry parameter at top of this file.
		*)
		
		function FillOutSmallArrayFromEntries(entries : Handle; numberOfEntries : longint; 
																		smallArray : SmallArrayHandle) : OSStatus;
			(* See comment where the routine is called. *)
			var
				err : OSStatus;
				entryIndex : longint;
				positionInEntryList : longint;
		begin
			err := noErr;		(* just in case numberOfEntries is 0 *)
			positionInEntryList := 0;
			for entryIndex := 1 to numberOfEntries do begin
				err := ICGetMapEntry(WhatGetDocumentInstance, entries, positionInEntryList, entry);
				if err <> noErr then begin
					leave;
				end; (* if *)
				smallArray^^[entryIndex].position := entryIndex;
				case item of
					ditFileMapSortExtension:
						smallArray^^[entryIndex].key := TPCopy(entry.extension, 1, 31);
					ditFileMapSortApplication:
						smallArray^^[entryIndex].key := TPCopy(entry.creator_app_name, 1, 31);
					ditFileMapSortName:
						smallArray^^[entryIndex].key := TPCopy(entry.entry_name, 1, 31);
				end; (* case *)
				positionInEntryList := positionInEntryList + entry.total_length;
			end; (* for *)
			FillOutSmallArrayFromEntries := err;
		end; (* FillOutSmallArrayFromEntries *)

		function BuildSortedEntriesFromSmallArray(entries : Handle; numberOfEntries : longint; 
											smallArray : SmallArrayHandle; sortedEntries : Handle) : OSStatus;
			(* See comment where the routine is called. *)
			var
				err : OSStatus;
				entryIndex : longint;
				positionInEntryList : longint;
				s : SInt8;
		begin
			err := noErr;		(* just in case numberOfEntries is 0 *)
			s := HGetState(entries);
			HLock(entries);
			for entryIndex := 1 to numberOfEntries do begin
				err := ICGetIndMapEntry(WhatGetDocumentInstance, entries, 
															smallArray^^[entryIndex].position, positionInEntryList, entry);
				if err = noErr then begin
					err := PtrAndHand(longint(entries^) + positionInEntryList, sortedEntries, entry.total_length);
				end; (* if *)
				if err <> noErr then begin
					leave;
				end; (* if *)
			end; (* for *)
			HSetState(entries, s);
			BuildSortedEntriesFromSmallArray := err;
		end; (* BuildSortedEntriesFromSmallArray *)
		
		var
			err : OSStatus;
			entries : Handle;
			sortedEntries:Handle;
			smallArray:SmallArrayHandle;
			numberOfEntries:longint;
	begin
		(* Switch to the watch cursor, 'cause this is going to take some time. *)
		SetCursor(GetCursor(watchCursor)^^);
		
		(* Get the current ICMapEntry list. *)
		entries := Handle(gWindows[wt].items[ditFileMapList]^.whatData);
		
		(* Create a destination for the sorted list. *)
		sortedEntries := NewHandle(0);
		err := MemError;
		
		(* Count the number of entries in the list and create a matching smallArray. *)
		smallArray := nil;		(* Prepare for failure. *)
		if err = noErr then begin
			err := ICCountMapEntries(WhatGetDocumentInstance, entries, numberOfEntries);
		end; (* if *)
		if err = noErr then begin
			smallArray := SmallArrayHandle(NewHandle(sizeof(SmallRecord) * numberOfEntries));
			err := MemError;
		end; (* if *)

		(* Walk the list of entries, filling out smallArray with the relevant key string. *)
		if err = noErr then begin
			err := FillOutSmallArrayFromEntries(entries, numberOfEntries, smallArray);
		end; (* if *)
		
		(* Sort the smallArray. *)
		if err = noErr then begin
			err := SmallSort(smallArray, numberOfEntries);
		end; (* if *)

		(* Now build sortedEntries by using the sorted smallArray to fetch the
			appropriate items from the original entries list and then tacking
			them on to the end of sortedEntries.
		*)
		if err = noErr then begin
			err := BuildSortedEntriesFromSmallArray(entries, numberOfEntries, smallArray,
																				sortedEntries);
		end; (* if *)
		
		(* Install the sortedEntries as the new global entries list.
			Note that SetEntriesHandle will dispose of a set of entries,
			one way or the other. If it succeeds, it'll dispose of the original
			entries list, and install sortedEntries in its place. If it fails,
			it'll dispose of sortedEntries. Either way, we must nil out
			sortedEntries here, so we don't dispose of it in the clean up code.
		*)
		if err = noErr then begin
			err := SetEntriesHandle(wt, sortedEntries);
			sortedEntries := nil;
		end; (* if *)
		
		(* Clean up, then switch back to the pointer cursor, and let the user know if
			we've had any problems.
		*)
		if sortedEntries <> nil then begin
			DisposeHandle(sortedEntries);
		end; (* if *)
		if smallArray <> nil then begin
			DisposeHandle(Handle(smallArray));
		end; (* if *)
		InitCursor;
		DisplayError(acSort, err);
	end; (* DoSortButtons *)
	
	(* ***** Defaults Button ***** *)
	
	procedure DoDefaultsButtons(wt : WindowType);
		(* Implements the functionality of the Defaults button in the File Mapping
			dialog. It simply gets the default mapping database from the application's
			resource fork, and puts it into the global entries list.
		*)
		var
			err : OSStatus;
			defaultEntries : Handle;
	begin
		(* Confirm with user first. *)
		if CautionAlert(rConfirmSetDefaultAlert, gOKCancelModalFilter) = ditOK then begin
		
			(* Get the default mapping database from the application's resource fork,
				and install it into the global entries list.
			*)
			err := GetDefaultMapEntries(defaultEntries);
			if err = noErr then begin
				err := SetEntriesHandle(wt, defaultEntries);
			end; (* if *)
			
			(* Report any errors, and clean up. *)
			DisplayError(acSetDefaults, err);
			if err = noErr then begin
				WhatDirtyDocument;
			end; (* if *)
		end; (* if *)
	end; (* DoDefaultsButtons *)

	(* ***** Delete Button ***** *)

	function DoDeleteButton(itemToDelete: integer; var entry: ICMapEntry;	
															entries: Handle; listH: ListHandle) : OSStatus;
		(* Implements the functionality of the Delete button in the File Mapping
			dialog. itemToDelete is the zero-based index of the item to delete.
			We just call IC to remove it from the entries list, and call List Manager
			to remove it from the List Handle. Not very tricky.
			See comment about entry parameter at top of this file.
		*)
		var
			err : OSStatus;
			positionOfEntryToDelete : longint;
	begin
		err := ICGetIndMapEntry(WhatGetDocumentInstance, entries, itemToDelete + 1, 
												positionOfEntryToDelete, entry);
		if err = noErr then begin
			err := ICDeleteMapEntry(WhatGetDocumentInstance, entries, positionOfEntryToDelete);
		end; (* if *)
		if err = noErr then begin
			LDelRow(1, itemToDelete, listH);
		end; (* if *)
		DoDeleteButton := err;
	end; (* DoDeleteButton *)

	(* ***** Layer Breaking Stuff ***** *)

	procedure MappingSetSelection(key:Str255);
		(* See comment in interface part. *)
		var
			err : OSStatus;
			window : WindowPtr;
			entries : Handle;
			listH : ListHandle;
			numberOfEntries : longint;
			entryIndex : longint;
			positionInEntryList : longint;
			entry : ICMapEntry;
	begin
		window := gWindows[WT_FileMapping].window;
		if window <> nil then begin
			(* Get the important global data. *)
			entries := Handle(gWindows[WT_FileMapping].items[ditFileMapList]^.whatData);
			listH := ListHandle(gWindows[WT_FileMapping].items[ditFileMapList]^.whatSpareData);
			err := ICCountMapEntries(WhatGetDocumentInstance, entries, numberOfEntries);
			
			(* Walk the list of entries, looking for the one whose extension matches the key. *)
			if err = noErr then begin
				positionInEntryList := 0;
				for entryIndex:= 1 to numberOfEntries do begin
					err := ICGetMapEntry(WhatGetDocumentInstance, entries, positionInEntryList, entry);
					if err <> noErr then begin
						leave;
					end; (* if *)
					
					(* If we have a matching key, open the Add/Edit Map Entry dialog
						with that entry and then bail out.
					*)
					if EqualString(entry.extension, key, false, true) then begin
						LSetSingleSelection(listH, entryIndex - 1);
						DimButtons(WT_FileMapping);
						if GetDControlEnable(window, ditFileMapChange) then begin
							if DoAddChangeButtons(entryIndex - 1, entry, entries, listH) = noErr then begin
								WhatDirtyDocument;
							end; (* if *)
						end; (* if *)
						leave;
					end; (* if *)
					
					positionInEntryList := positionInEntryList + entry.total_length;
				end; (* for *)
			end; (* if *)
		end; (* if *)
	end; (* MappingSetSelection *)

	(* ***** Standard What Interface ***** *)

	(* ***** What Init ***** *)
	function WhatInitFileMap(dummy : integer) : OSStatus;
	begin
		{$unused dummy}
		gListUserItemUPP := NewUserItemProc(@ListUserItemUpdateProc);
		gFileMapLDEFUPP := NewListDefProc(@FileMapLDEF);
		WhatInitFileMap := noErr;
	end; (* WhatInitFileMap *)

	(* ***** What Open ***** *)

	function WhatOpenFileMap (wt: WindowType; item: integer): OSStatus;
		(* See the comments in ICWindowGlobals. *)
		var
			err: OSStatus;
			window: DialogPtr;
			entries: Handle;
			attr: longint;
			numberOfEntries: longint;
			listItemRect: Rect;
			listDataRect: Rect;
			listCellSize: Point;
			listH: ListHandle;
	begin
		(* Prepare for failure. *)
		gWindows[wt].items[item]^.whatSpareData := nil;
		gWindows[wt].items[item]^.whatData := nil;
		err := noErr;
		
		(* Install the user item update proc for the list item. *)
		window := gWindows[wt].window;
		SetDItemHandle(window, item, gListUserItemUPP);
		
		(* Get the list of entries, if we fail then create an empty one. *)
		if ICGetPrefHandle(WhatGetDocumentInstance, gWindows[wt].items[item]^.whatKey, attr, entries) <> noErr then begin
			entries := NewHandle(0);
			if entries = nil then begin
				err := memFullErr;
			end; (* if *)
		end; (* if *)
		WhatProcessAttributes(wt, item, attr);
		
		(* Create the ListHandle for the list item. The list starts out one cell wide
			and numberOfEntries cells high. The horizontal cell size is governed by
			the width of the list user item and kMappingCellHeight. The total size of the list
			is determined by the list user item, minus the width of the scroll bar.
		*)
		if err = noErr then begin
			if ICCountMapEntries(WhatGetDocumentInstance, entries, numberOfEntries) <> noErr then begin
				numberOfEntries := 0;
			end; (* if *)
			gWindows[wt].items[item]^.whatData := Ptr(entries);
			GetDItemRect(window, item, listItemRect);
			listItemRect.right := listItemRect.right - 15;
			SetRect(listDataRect, 0, 0, 1, numberOfEntries);
			SetPt(listCellSize, listItemRect.right - listItemRect.left, kMappingCellHeight);
			listH := LNew(listItemRect, listDataRect, listCellSize, rSimpleLDEF, window, 
											true, false, false, true);
			if listH = nil then begin
				err := memFullErr;
			end; (* if *)
		end; (* if *)
		
		(* All is cool, finish up the open. *)
		if err = noErr then begin
			(* Record the ListHandle for the other routines. *)
			gWindows[wt].items[item]^.whatSpareData := Ptr(listH);
			(* Set the refCon of the ListHandle so that the Simple LDEF calls FileMapLDEF. *)
			listH^^.refCon := ord(gFileMapLDEFUPP);
			(* Set the list flags, so we only get one item selected. *)
			listH^^.selFlags := lOnlyOne;
			(* Enable/disable the various buttons. *)
			DimButtons(wt);
			(* Install the user item proc to draw the thick rounded rect around the default button. *)
			SetupDefaultButtonUserItem(window, ditFileMapChange, ditFileMapDefaultUserItem);
			(* Ensure that we see any keystrokes typed into this window. *)
			gWindows[wt].selectedItem := ditFileMapList;
		end; (* if *)
		
		WhatOpenFileMap := err;
	end; (* WhatOpenFileMap *)

	(* ***** What Flush ***** *)

	function WhatFlushFileMap (wt: WindowType; item: integer): OSStatus;
		(* See the comments in ICWindowGlobals. *)
		var
			err: OSStatus;
	begin
		if WhatIsLocked(wt, item) then begin
			err := noErr;
		end else begin
			err := ICSetPrefHandle(WhatGetDocumentInstance, gWindows[wt].items[item]^.whatKey, ICattr_no_change, 
															Handle(gWindows[wt].items[item]^.whatData));
		end; (* if *)
		WhatFlushFileMap := err;
	end; (* WhatFlushFileMap *)

	(* ***** What Close ***** *)

	function WhatCloseFileMap (wt: WindowType; item: integer): OSStatus;
		(* See the comments in ICWindowGlobals. *)
	begin
		if gWindows[wt].items[item]^.whatSpareData <> nil then begin
			LDispose(ListHandle(gWindows[wt].items[item]^.whatSpareData));
		end; (* if *)
		if gWindows[wt].items[item]^.whatData <> nil then begin
			DisposeHandle(Handle(gWindows[wt].items[item]^.whatData));
		end; (* if *)
		WhatCloseFileMap := noErr;
	end; (* WhatCloseFileMap *)

	(* ***** What Key ***** *)

	function WhatKeyFileMap (wt: WindowType; item: integer; event: EventRecord): OSStatus;
		(* See the comments in ICWindowGlobals. *)
		var
			err: OSStatus;
			listH: ListHandle;
			entry: ICMapEntry;
	begin
		{$unused item}
		listH := ListHandle(gWindows[wt].items[ditFileMapList]^.whatSpareData);
		err := noErr;
		if (event.what = keyDown) or (event.what = autoKey) then begin
			case chr(band(event.message, charCodeMask)) of
				kCRChar, kEnterChar:  begin
					(* Return and Enter trigger the Change button. *)
					if GetDControlEnable(gWindows[wt].window, ditFileMapChange) then begin
						FlashDItem(gWindows[wt].window, ditFileMapChange);
						err := DoAddChangeButtons(LSelectedLine(listH), entry, Handle(gWindows[wt].items[ditFileMapList]^.whatData), listH);
						if err = noErr then begin
							WhatDirtyDocument;
						end; (* if *)
					end; (* if *)
				end;
				otherwise begin
					(* Change the selection based on the typed character. *)
					LDoKey(listH, event, GetEntryName);
					DimButtons(wt);
				end;
			end; (* case *)
		end; (* if *)
		WhatKeyFileMap := err;
	end; (* WhatKeyFileMap *)

	(* ***** What Click ***** *)
	
	function WhatClickFileMap (wt: WindowType; item: integer; event: EventRecord): OSStatus;
		(* See comment in interface part. *)
		var
			err: OSStatus;
			entries: Handle;
			listH: ListHandle;
			selectedListCell: integer;
			entry: ICMapEntry;
	begin
		(* General preparation. *)
		entries := Handle(gWindows[wt].items[ditFileMapList]^.whatData);
		listH := ListHandle(gWindows[wt].items[ditFileMapList]^.whatSpareData);
		selectedListCell := LSelectedLine(listH);
		err := 1;
		
		(* Dispatch for each item in the dialog. *)
		case item of
			ditFileMapAdd: 
				err := DoAddChangeButtons(-1, entry, entries, listH);
			ditFileMapChange: 
				err := DoAddChangeButtons(selectedListCell, entry, entries, listH);
			ditFileMapSortExtension,
			ditFileMapSortApplication,
			ditFileMapSortName:
				DoSortButtons(wt, item, entry);
			ditFileMapDefaults:
				DoDefaultsButtons(wt);
			ditFileMapDelete: 
				if selectedListCell <> -1 then begin
					err := DoDeleteButton(selectedListCell, entry, entries, listH);
				end; (* if *)
			ditFileMapList:  begin
				(* A click in the main list item. Call through to List Manager's LClick
					routine, and then attempt to open the Add/Edit Mapping dialog if
					it indicates a double click.
				*)
				GlobalToLocal(event.where);
				if LClick(event.where, 0, listH) then begin
					if WhatIsLocked(wt, item) then begin
						WhatLockedAlert(wt, item);
						err := userCanceledErr;
					end else begin
						FlashDItem(gWindows[wt].window, ditFileMapChange);
						(* Call LSelectedLine again beacuse selection may have been changed by LClick. *)
						err := DoAddChangeButtons(LSelectedLine(listH), entry, entries, listH);
					end; (* if *)
				end; (* if *)
			end;
		end; (* case *)
		
		(* Clean up. *)
		if err = noErr then begin
			WhatDirtyDocument;
		end; (* if *)
		if err = 1 then begin
			err := noErr;
		end; (* if *)
		DimButtons(wt);
		WhatClickFileMap := err;
	end; (* WhatClickFileMap *)

	(* ***** What Activate ***** *)

	function WhatActivateFileMap (wt: WindowType; item: integer; activate: Boolean): OSStatus;
		(* See the comments in ICWindowGlobals. *)
	begin
		(* Pass the event straight through to List Manager. *)
		LActivate(activate, ListHandle(gWindows[wt].items[item]^.whatSpareData));
		WhatActivateFileMap := noErr;
	end; (* WhatActivateFileMap *)

	(* ***** What Cursor ***** *)

	function WhatCursorFileMap (wt: WindowType; item: integer; pt: Point; cursorID: integer): OSStatus;
		(* See the comments in ICWindowGlobals and ICWhatSubs. *)
	begin
		WhatCursorFileMap := WhatCursorListItem(wt, item, pt, cursorID);
	end; (* WhatCursorFileMap *)

end. (* ICFileMapWhat *)
