unit ICRGBColourWhats;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This module implements the RGB Colour What, a simple user interface that allows
	for colour selection.  Note that this is yet another What that's responsible for
	more than one dialog item.
	
		Item#		Item Type		What				Purpose
		-----		---------		----				-------
		item			user item		n/a				display colour sample
		item + 1	button			RGBColour	brings up Colour Picker to change colour

	The whatData field of the RGB Colour What holds a pointer to the currently
	selected colour.  If that pointer is nil, the system doesn't support colour.
*)

interface

	uses
		ICWindowGlobals;

	(* ***** Standard What Interface ***** *)
	
	(* See the comments in ICWindowGlobals. *)

	function WhatInitRGBColour(dummy : integer) : OSStatus;
	function WhatOpenRGBColour (wt: WindowType; item: integer): OSStatus;
	function WhatFlushRGBColour (wt: WindowType; item: integer): OSStatus;
	function WhatCloseRGBColour (wt: WindowType; item: integer): OSStatus;
	function WhatClickRGBColour (wt: WindowType; item: integer; event: EventRecord): OSStatus;

implementation

	uses
		ColorPicker,
		
		InternetConfig,
		ICStringLists,
		ICDialogs,
		ICWindowGlobals,
		ICMiscSubs,
		ICResources,
		ICGlobals;

	var
		gRGBSampleUserItemUpdateUPP : UserItemUPP;

	(* ***** Drawing ***** *)

	procedure RGBSampleUserItemUpdateProc (window: DialogPtr; item: integer);
		(* This routine is called by the Dialog Manager to redraw the user
			item that displays the sample colour.  Note the subtle test
			for currentColour being nil.  WhatOpenRGBColour sets currentColour
			to nil if we don't have Colour QuickDraw.  The test here prevents
			this routine from calling colour routines on a B&W Mac.
		*)
		var
			wt : WindowType;
			oldPenState : PenState;
			itemRect : Rect;
			currentColour : RGBColorPtr;
			oldColour : RGBColor;
	begin
		wt := MapWindowPtrToWindowType(window);
		GetPenState(oldPenState);
		PenNormal;
		GetDItemRect(window, item, itemRect);
		FrameRect(itemRect);
		currentColour := RGBColorPtr(gWindows[wt].items[item + 1]^.whatData);
		if currentColour <> nil then begin
			GetForeColor(oldColour);
			RGBForeColor(currentColour^);
			InsetRect(itemRect, 3, 3);
			PaintRect(itemRect);
			RGBForeColor(oldColour);
		end; (* if *)
		SetPenState(oldPenState)
	end; (* RGBSampleUserItemUpdateProc *)

	(* ***** Standard What Interface ***** *)

	(* ***** What Init ***** *)

	function WhatInitRGBColour(dummy : integer) : OSStatus;
	begin
		{$unused dummy}
		gRGBSampleUserItemUpdateUPP := NewUserItemProc(@RGBSampleUserItemUpdateProc);
		WhatInitRGBColour := noErr;
	end; (* WhatInitRGBColour *)

	(* ***** What Open ***** *)

	function WhatOpenRGBColour (wt: WindowType; item: integer): OSStatus;
		(* See the comments in ICWindowGlobals. *)
		var
			prefSize : longint;
			currentColour : RGBColor;
			attr : ICAttr;
	begin
		(* Get the current setting from IC. *)
		prefSize := sizeof(RGBColor);
		if ICGetPref(WhatGetDocumentInstance, gWindows[wt].items[item]^.whatKey, attr, @currentColour, prefSize) <> noErr then begin
			currentColour.red := -1;			(* Pascal defines RGBColor using signed fields, duh!, so -1 = $FFFF = white *)
			currentColour.green := -1;
			currentColour.blue := -1;
		end; (* if *)
		WhatProcessAttributes(wt, item, attr);
		
		(* Create a pointer to the colour and store it in whatData. *)
		if gHasColourQuickDraw then begin
			gWindows[wt].items[item]^.whatData := NewPtr(sizeof(RGBColor));
			if gWindows[wt].items[item]^.whatData <> nil then begin
				RGBColorPtr(gWindows[wt].items[item]^.whatData)^ := currentColour;
			end; (* if *)
		end else begin
			(* We don't have Colour QuickDraw, so save a nil pointer, which
				tells RGBSampleUserItemUpdateProc not to do any drawing.
				Also disable the button, 'cause we can't change the setting
				with Colour QuickDraw's Colour Picker.
			*)
			gWindows[wt].items[item]^.whatData := nil;
			SetDControlEnable(gWindows[wt].window, item, false);
		end; (* if *)
		
		(* Finish up.  Disable the button if the setting is locked, note that we've
			not yet been modified, and install the update proc.
		*)
		if WhatIsLocked(wt, item) then begin
			SetDControlEnable(gWindows[wt].window, item, false);
		end; (* if *)
		gWindows[wt].items[item]^.whatModified := false;
		SetDItemHandle(gWindows[wt].window, item - 1, gRGBSampleUserItemUpdateUPP);

		WhatOpenRGBColour := noErr;
	end; (* WhatOpenRGBColour *)

	(* ***** What Flush ***** *)

	function WhatFlushRGBColour (wt: WindowType; item: integer): OSStatus;
		(* See the comments in ICWindowGlobals. *)
		var
			err : OSStatus;
			currentColour : RGBColorPtr;
	begin
		err := noErr;
		currentColour := RGBColorPtr(gWindows[wt].items[item]^.whatData);
		if (currentColour <> nil) & (gWindows[wt].items[item]^.whatModified) then begin
			err := ICSetPref(WhatGetDocumentInstance, gWindows[wt].items[item]^.whatKey, ICattr_no_change, 
																Ptr(currentColour), sizeof(RGBColor));
			gWindows[wt].items[item]^.whatModified := false;
		end; (* if *)
		WhatFlushRGBColour := err;
	end; (* WhatFlushRGBColour *)

	(* ***** What Close ***** *)

	function WhatCloseRGBColour (wt: WindowType; item: integer): OSStatus;
		(* See the comments in ICWindowGlobals. *)
	begin
		if gWindows[wt].items[item]^.whatData <> nil then begin
			DisposePtr(gWindows[wt].items[item]^.whatData);
			gWindows[wt].items[item]^.whatData := nil;
		end; (* if *)
		WhatCloseRGBColour := noErr;
	end; (* WhatCloseRGBColour *)

	(* ***** What Click ***** *)
	
	function WhatClickRGBColour (wt: WindowType; item: integer; event: EventRecord): OSStatus;
		(* See the comments in ICWindowGlobals. *)
		var
			currentColour : RGBColorPtr;
			chosenColour : RGBColor;
	begin
		{$unused event}
		currentColour := RGBColorPtr(gWindows[wt].items[item]^.whatData);
		if currentColour <> nil then begin
			if GetColor(Point(0), 
								GetIndStr(rMiscStrings, strSelectColourPrompt), 
								currentColour^, 
								chosenColour) then begin
				currentColour^ := chosenColour;
				InvalDItem(gWindows[wt].window, item - 1);
				gWindows[wt].items[item]^.whatModified := true;
				WhatDirtyDocument;
			end; (* if *)
		end; (* if *)
		WhatClickRGBColour := noErr;
	end; (* WhatClickRGBColour *)
	
end. (* ICRGBColourWhats *)
